@@ -16,6 +16,7 @@
 #include <sys/ioctl.h>
 #include <sys/stat.h>
 #include <utmp.h>
+#include <crypt.h>
 
 #include "../../ldm.h"
 #include "../../ldmutils.h"
@@ -32,8 +33,9 @@
 LdmBackend *descriptor;
 SSHInfo *sshinfo;
 
-void __attribute__((constructor)) initialize() {
-    descriptor = (LdmBackend*) malloc(sizeof(LdmBackend));
+void __attribute__ ((constructor)) initialize()
+{
+    descriptor = (LdmBackend *) malloc(sizeof(LdmBackend));
     bzero(descriptor, sizeof(LdmBackend));
 
     descriptor->name = "ssh";
@@ -50,8 +52,10 @@ void __attribute__((constructor)) initialize() {
  * init_ssh
  *  Callback function for initialization
  */
-void init_ssh() {
-    sshinfo = (SSHInfo*) malloc(sizeof(SSHInfo));
+void
+init_ssh()
+{
+    sshinfo = (SSHInfo *) malloc(sizeof(SSHInfo));
     bzero(sshinfo, sizeof(SSHInfo));
 
     /* Get ENV Variables */
@@ -63,22 +67,24 @@ void init_ssh() {
  * start_ssh
  *  Start ssh session
  */
-void start_ssh() {
+void
+start_ssh()
+{
     gboolean error = FALSE;
 
     /* Variable validation */
     if (!(sshinfo->username)) {
-        log_entry("ssh",3,"no username");
+        log_entry("ssh", 3, "no username");
         error = TRUE;
     }
 
     if (!(sshinfo->password)) {
-        log_entry("ssh",3,"no password");
+        log_entry("ssh", 3, "no password");
         error = TRUE;
     }
 
     if (!(sshinfo->server)) {
-        log_entry("ssh",3,"no server");
+        log_entry("ssh", 3, "no server");
         error = TRUE;
     }
 
@@ -86,7 +92,7 @@ void start_ssh() {
         sshinfo->session = g_strdup("default");
 
     if (error) {
-        die("ssh","missing mandatory information");
+        die("ssh", "missing mandatory information");
     }
 
     /* Getting Xsession */
@@ -99,25 +105,41 @@ void start_ssh() {
      * If we run multiple ldm sessions on multiply vty's we need separate
      * control sockets.
      */
-    sshinfo->ctl_socket = g_strdup_printf("/var/run/ldm_socket_%d_%s", ldm.pid, sshinfo->server);
+    sshinfo->ctl_socket =
+        g_strdup_printf("/var/run/ldm_socket_%d_%s", ldm.pid,
+                        sshinfo->server);
 
     /* Setting ENV variables for plugin */
     _set_env();
 
     /* Execute any rc files */
-    log_entry("ssh",6,"calling rc.d pressh scripts");
+    log_entry("ssh", 6, "calling rc.d pressh scripts");
     rc_files("pressh");
 
     ssh_session();
-    log_entry("ssh",6,"established ssh session on '%s' as '%s'",sshinfo->server,sshinfo->username);
+    log_entry("ssh", 6, "established ssh session on '%s' as '%s'",
+              sshinfo->server, sshinfo->username);
 
     /* Greeter not needed anymore */
     close_greeter();
 
-    log_entry("ssh",6,"calling rc.d start scripts");
-    rc_files("start");                      /* Execute any rc files */
+    log_entry("ssh", 6, "calling rc.d start scripts");
+    rc_files("start");                           /* Execute any rc files */
 
-    log_entry("ssh",6,"starting X session");
+    /* ssh_hashpass - Defaults to opt-in (Must set LDM_PASSWORD_HASH to true) */
+    if (ldm_getenv_bool_default("LDM_PASSWORD_HASH", FALSE))    
+    {    
+        ssh_hashpass();
+    } 
+    else
+    {
+        log_entry("hashpass", 6, "LDM_PASSWORD_HASH set to FALSE or unset, skipping hash function");
+    }        
+    log_entry("hashpass", 6, "Freeing password as promised.");
+    g_free(sshinfo->password);
+    sshinfo->password = NULL;
+
+    log_entry("ssh", 6, "starting X session");
     set_session_env(sshinfo->xsession, sshinfo->session);
 }
 
@@ -125,8 +147,10 @@ void start_ssh() {
  * get_guest
  *  Callback function for setting guest login
  */
-void get_guest() {
-    log_entry("ssh",6,"setting guest login");
+void
+get_guest()
+{
+    log_entry("ssh", 6, "setting guest login");
 
     /* Get credentials */
     g_free(sshinfo->username);
@@ -178,11 +202,11 @@ void get_guest() {
         good = FALSE;
         if (sshinfo->server) {
             i = 0;
-            while(1) {
-                if(hosts_char[i] == NULL) {
+            while (1) {
+                if (hosts_char[i] == NULL) {
                     break;
                 }
-                if(!g_strcmp0(hosts_char[i], sshinfo->server)) {
+                if (!g_strcmp0(hosts_char[i], sshinfo->server)) {
                     good = TRUE;
                     break;
                 }
@@ -203,7 +227,9 @@ void get_guest() {
  * _set_env
  *  Set environment variables used by LDM and Greeter
  */
-void _set_env() {
+void
+_set_env()
+{
     setenv("LDM_SERVER", sshinfo->server, 1);
     setenv("LDM_USERNAME", sshinfo->username, 1);
     setenv("LDM_SOCKET", sshinfo->ctl_socket, 1);
@@ -213,7 +239,9 @@ void _set_env() {
  * get_auth
  *  Callback function for authentication
  */
-void get_auth() {
+void
+get_auth()
+{
     /* Get UserID */
     get_userid(&(sshinfo->username));
 
@@ -234,8 +262,10 @@ void get_auth() {
  * close_ssh
  *  Callback function for closing the plugins
  */
-void close_ssh() {
-    log_entry("ssh",7,"closing ssh session");
+void
+close_ssh()
+{
+    log_entry("ssh", 7, "closing ssh session");
     ssh_endsession();
 
     // leave no crumbs and free memory allocated for auth values
@@ -247,7 +277,9 @@ void close_ssh() {
     free(sshinfo);
 }
 
-int expect(int fd, char *p, int seconds, ...) {
+int
+expect(int fd, char *p, int seconds, ...)
+{
     fd_set set;
     struct timeval timeout;
     int i, st;
@@ -281,8 +313,8 @@ int expect(int fd, char *p, int seconds, ...) {
      * Main loop.
      */
 
-    while(1) {
-        timeout.tv_sec = (long)1;             /* one second timeout */
+    while (1) {
+        timeout.tv_sec = (long) 1;               /* one second timeout */
         timeout.tv_usec = 0;
 
         FD_ZERO(&set);
@@ -290,10 +322,10 @@ int expect(int fd, char *p, int seconds, ...) {
         st = select(FD_SETSIZE, &set, NULL, NULL, &timeout);
 
         if (child_exited) {
-            break;                  /* someone died on us */
+            break;                               /* someone died on us */
         }
 
-        if (st < 0) {                 /* bad thing */
+        if (st < 0) {                            /* bad thing */
             break;
         }
 
@@ -301,8 +333,8 @@ int expect(int fd, char *p, int seconds, ...) {
             break;
         }
 
-        if (!st) {                  /* timeout */
-            loopcount--;            /* We've not seen the data we want */
+        if (!st) {                               /* timeout */
+            loopcount--;                         /* We've not seen the data we want */
             continue;
         }
 
@@ -332,19 +364,21 @@ int expect(int fd, char *p, int seconds, ...) {
         }
     }
 
-    log_entry("ldm",7,"expect saw: %s", p);
+    log_entry("ldm", 7, "expect saw: %s", p);
 
     if (size < 0 || st < 0 || child_exited) {
-        return ERROR;               /* error occured */
+        return ERROR;                            /* error occured */
     }
     if (loopcount == 0) {
-        return TIMED_OUT;           /* timed out */
+        return TIMED_OUT;                        /* timed out */
     } else {
-        return i;                   /* which expect did we see? */
+        return i;                                /* which expect did we see? */
     }
 }
 
-void ssh_chat(gint fd) {
+void
+ssh_chat(gint fd)
+{
     int seen;
     gchar lastseen[MAXEXP];
     int first_time = 1;
@@ -364,8 +398,6 @@ void ssh_chat(gint fd) {
         /* We might have a : in the data, we're looking for :'s at the
            end of the line */
         if (seen == 0) {
-            g_free(sshinfo->password);
-            sshinfo->password = NULL;
             return;
         } else if (seen == 1) {
             int i;
@@ -388,15 +420,17 @@ void ssh_chat(gint fd) {
             set_message(_("No response from server, restarting..."));
             sleep(5);
             close_greeter();
-            die("ssh","no response, restarting");
+            die("ssh", "no response, restarting");
         }
     }
 }
 
-void ssh_tty_init(void) {
+void
+ssh_tty_init(void)
+{
     (void) setsid();
-    if (login_tty(sshinfo->sshslavefd) < 0 ) {
-        die("ssh","login_tty failed");
+    if (login_tty(sshinfo->sshslavefd) < 0) {
+        die("ssh", "login_tty failed");
     }
 }
 
@@ -404,25 +438,28 @@ void ssh_tty_init(void) {
  * ssh_session()
  * Start an ssh login to the server.
  */
-void ssh_session(void) {
+void
+ssh_session(void)
+{
     gchar *command;
     gchar *port = NULL;
     pthread_t pt;
 
     /* Check for port Override */
-    if(sshinfo->override_port)
-        port = g_strconcat(" -p ", sshinfo->override_port, " ",  NULL);
+    if (sshinfo->override_port)
+        port = g_strconcat(" -p ", sshinfo->override_port, " ", NULL);
 
     openpty(&(sshinfo->sshfd), &(sshinfo->sshslavefd), NULL, NULL, NULL);
 
     command = g_strjoin(" ", "ssh", "-Y", "-t", "-M",
-            "-S", sshinfo->ctl_socket,
-            "-o", "NumberOfPasswordPrompts=1",
-            "-l", sshinfo->username,
-            port ? port : "",
-            sshinfo->sshoptions ? sshinfo->sshoptions : "",
-            sshinfo->server, "echo " SENTINEL "; exec /bin/sh -", NULL);
-    log_entry("ssh",7,"ssh_session: %s", command);
+                        "-S", sshinfo->ctl_socket,
+                        "-o", "NumberOfPasswordPrompts=1",
+                        "-l", sshinfo->username,
+                        port ? port : "",
+                        sshinfo->sshoptions ? sshinfo->sshoptions : "",
+                        sshinfo->server,
+                        "echo " SENTINEL "; exec /bin/sh -", NULL);
+    log_entry("ssh", 7, "ssh_session: %s", command);
 
     sshinfo->sshpid = ldm_spawn(command, NULL, NULL, ssh_tty_init);
 
@@ -437,7 +474,9 @@ void ssh_session(void) {
         g_free(port);
 }
 
-void ssh_endsession(void) {
+void
+ssh_endsession(void)
+{
     GPid pid;
     gchar *command;
     struct stat stbuf;
@@ -445,8 +484,10 @@ void ssh_endsession(void) {
     if (!stat(sshinfo->ctl_socket, &stbuf)) {
         /* socket still exists, so we need to shut down the ssh link */
 
-        command = g_strjoin(" ", "ssh", "-S", sshinfo->ctl_socket, "-O", "exit", sshinfo->server, NULL);
-        log_entry("ssh",7,"closing ssh session: %s"), command;
+        command =
+            g_strjoin(" ", "ssh", "-S", sshinfo->ctl_socket, "-O", "exit",
+                      sshinfo->server, NULL);
+        log_entry("ssh", 7, "closing ssh session: %s"), command;
         pid = ldm_spawn(command, NULL, NULL, NULL);
         ldm_wait(pid);
         close(sshinfo->sshfd);
@@ -456,7 +497,63 @@ void ssh_endsession(void) {
     }
 }
 
-void *eater() {
+/*
+ * ssh_hashpass()
+ * Set up password hash for client /etc/shadow using /dev/urandom
+ * rather than g_rand() due to developer recommendations at:
+ * https://developer.gnome.org/glib/stable/glib-Random-Numbers.html
+ */
+void
+ssh_hashpass(void)
+{
+    FILE *rand_fp;
+    FILE *shad_fp;
+    gchar salt[] = "$6$...............$";
+    gchar buf[16];
+    const gchar seedchars[] =
+        "./0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";
+    gchar *shadowentry;
+    const gchar hashloc[] = "/var/cache/ltsp/shadow.sed";
+    size_t i = 0;
+    log_entry("hashpass", 6, "LDM_PASSWORD_HASH set to true, setting hash");        
+    rand_fp = fopen("/dev/urandom", "r");
+    if (rand_fp == NULL) 
+    {
+        log_entry("hashpass", 7, "Unable to read from /dev/urandom - Skipping HASH function");
+    } 
+    else 
+    {     
+        fread(buf, sizeof buf, 1, rand_fp);
+        fclose(rand_fp);
+        for (; i < sizeof buf; i++) 
+        {
+            salt[3 + i] = seedchars[buf[i] % (sizeof seedchars - 1)];
+        }
+        shadowentry = crypt(sshinfo->password, salt);
+        log_entry("hashpass", 6, "hash created");
+        /* generate dynamic file for writing hash to. 
+        * Will remove anything in its way. 
+        * This will be removed during rc.d script run.
+        */
+        shad_fp = fopen(hashloc, "w");
+        if (shad_fp == NULL) 
+        {
+            log_entry("hashpass", 7, "Unable to open %s for hash entry.",
+                      hashloc);
+        }
+        else
+        {
+            fprintf(shad_fp,
+                    "# Generated by LTSP, for LDM rc.d script manipulation\n$s:!:%s:",
+                    shadowentry);
+            fclose(shad_fp);
+        }
+    } 
+}
+
+void *
+eater()
+{
     fd_set set;
     struct timeval timeout;
     int st;
@@ -468,7 +565,7 @@ void *eater() {
             break;
         }
 
-        timeout.tv_sec = (long)1;             /* one second timeout */
+        timeout.tv_sec = (long) 1;               /* one second timeout */
         timeout.tv_usec = 0;
         FD_ZERO(&set);
         FD_SET(sshinfo->sshfd, &set);
